
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                mainSOBI.m                                      %
%                                                                %
%pattern after                                                   %
%Rainieri, C.,                                                   %
%Perspectives of Second-Order Blind Identification               %  
%for Operational Modal Analysis of Civil Structures.             %
%Shock and Vibration, 2014.                                      %
%                                                                %
%The Joint Diagonalization is provided by Jean-Francois Cardoso  %
%and is available from:                                          %
%http://perso.telecom-paristech.fr/~cardoso/jointdiag.html.      %
%                                                                %
%Implemented by Ruth Hetland, Trondheim, June 2015               %
%                                                                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 clc; close all; clear; % reset workspace

%% INPUT 
%  Load an lxn dimensional matrix, where l is the number 
%  of signals and n the number of sampling points

prompt='Load, then press enter';
data=input(prompt)    
ID='00'  %Number the save files with an ID
ddu=data.framedata; %save the lxn dimensional matrix to ddu variable

 %% Define variables
global fs l Nm n   %make variables global
    
fs=100;     %sampling frequency, Hz
Nm=4;       %expected number of activemodes, for classical whitening only
K=1000;     %number of time lags
thresh=1e-8;%treshold value, for use in the JAD 


%% Signal processing
y_f = detrend(ddu','constant')'; % remove offset by detrending the dataset

%create filter
fch=1.5; fcl=0.2;  % define higer/lower cut of frequency 
fch_norm=2*fch/fs; fcl_norm=2*fcl/fs;%create normalized cutoff frequencies 
[b3 a3] = butter(5, fch_norm, 'low'); %create Butterworth filter, lowpass
[b3_h a3_h] = butter(5, fcl_norm, 'high'); %create Butterworth filter, highpass

%use zero-phase digital filtering to apply Butterworth filter in time domain
y_f=filtfilt(b3,a3,y_f')'; 
y_f=filtfilt(b3_h,a3_h,y_f')';

%resampling of dataserie, detrend
f1=1; f2=5; %factors to resample
y=resample(y_f',f1,f2)'; 
fs=fs*f1/f2; %new sampling frequency

%% Datasetproperties
[l n]=size(y); %l=number of signals  n=columns observed data
t=0:1/fs:(n-1)/fs; %time vector

%% Calculate whitening matrix
%robust whitening input: detrended lxn matrix y. with l time series and n
%samples. Output a whitening matrix Wr with size lxl=NmxNm 
%Wrp=pseudoinvers
%classical whitening input: y=detended signal, K=number of timelags
%In addition, the classical_whitening function includes calculation 
%of correlation matrix of the signals
[Wr Wrp RR]=classical_whitening(y,K);
for i=1:length(t)
z_y(:,i)=Wr*y(:,i);
end
%%Check that Rzz(0) equals identity matrix
Rzz_0=cov(z_y');

%% U=unitary matrix obtained as joint diagonalizer of the set of ccovariances
%for a fixed set of time lags
% Joint approximate diagonalization of n matrices of size m*m stored in the
% m*mn matrix A by minimization of a joint diagonality criterion
[ Arcurl , Rcl ] =  rjd(RR,thresh);
Ur=Arcurl'*Wr; %demixing matrix U
Ar=Wrp*Arcurl; %mxing matrix
%obtain sourcces
for i=1:length(t)
sr(:,i)=Ur*y(:,i); %source signal estimated from the demixing matrix
end

%% plot after JADE
c=1; %count lag

[Ryy_cov_sr lags]=xcov(sr',K,'unbiased')
Ryy_sr=Ryy_cov_sr(K+2:2*K+1,:);
Lsr=lags([K+2:2*K+1]);
nlsr=length(Lsr); %in case dK~=1;

figure
for i=1:Nm*Nm
        subplot(Nm,Nm,i)
        hold on
        plot(1:K,Ryy_sr(:,i)) 
        axis([0 K -1 1])
end
suptitle('R(\tau) after joint diagonalization')
savefig(['ID' num2str(ID) 'after_joint_diagonalization'])
%% plot sources/PSD 
figure(100)
suptitle('Identified sources')% found with robust whitening')
i=1;
segments=5;
window=round(n/segments);
noverlap=window/2;


for j=1:2:Nm*2
    subplot(Nm,2,j)
        plot(t,sr(i,:))
        ylabel('s(t)')
        axis([0 t(60*(fs)) -1.1*max(abs(sr(i,:))) 1.1*max(abs(sr(i,:)))])
        
        [PSD_sr{i} f{i}]=pwelch(sr(i,:)',window,noverlap,4*window,fs);
        [pf{i} fre{i}]=findpeaks(PSD_sr{i});
        [ppf frq]=max(pf{i});
        f_sr(i)=f{i}(fre{i}(frq));

    subplot(Nm,2,j+1)
        plot( f{i},PSD_sr{i},'b')
        hold on
        plot(f{i}(fre{i}(frq)),ppf,'+r')
        text(2,ppf,[num2str(f{i}(fre{i}(frq)),3) ' Hz'],'color','r')
        ylabel(['PSD'])
        xlabel(['frequency [Hz]'])
        axis([0 1.5 0 1.5*max(PSD_sr{i})])
i=i+1;
end
savefig(['ID' num2str(ID) 'identifiedSources'])

%% damping and frequency found from correlation functions 
  %find the autocorrelation of sources,
  %the correlations are provided from the rjd function
nR=length(Rcl)
for i=1:Nm
    for ii=1:nR/Nm
        j=i+(ii-1)*Nm;
        Rrss(i,ii)=(Rcl(i,j));
    end
end 
%% PLOT correlation functions of sources
figure(200)
suptitle('Correlation functions of sources')
for i=1:Nm
    subplot(Nm,1,i)
plot(1:K,Rrss(i,:))
ylabel(['R_s_' num2str(i) '_s_' num2str(i)])
if i==Nm
xlabel('\tau')
end

end
savefig(['ID' num2str(ID) 'correlation function'])

%% input number of timelags for zerocrossing
for j=1:Nm
    prompt=['includetimelags, ' num2str(j) ': '];
    zc(j)=input(prompt)
    [upeak{j},tpeak{j}]=findpeaks(abs(real(Rrss(j,:)))); % Find peaks 
    [f_R x0]=zerocross(real(Rrss(j,1:zc(j))),1:zc(j));
    figure(200)
        subplot(Nm,1,j)
            hold on
        z0x0=zeros(1,length(x0{1}));
        plot(x0{1},z0x0,'--or')
        plot(tpeak{j},upeak{j},'og')
        clear x0
        f_Rrss(j)=f_R;
end
%% Logaritmic decrement %%input number of peaks
s=[1 1 1 1 1 1]; %startpeaknumber
%calculate logarimic decrement from maximas
for j=1:Nm
    c=1;
    for k=1:length(upeak{j})-s(j);
        dec_y{j}(c)=2*log(abs(upeak{j}(s(j)))/abs(upeak{j}(k+s(j))));
        c=c+1;
    end
end

%specify number of maximas
for j=1:Nm
prompt=['numberofpeaks' num2str(j) ': '];
nofpeaks(j)=input(prompt)
end

%linear regression with the specified number of maximas
reg_line=[]; Pp=[];
figure
suptitle('Linear relationship of logaritmic decrement')
for j=1:Nm
    Pp=[];
    pk{j}=1:(nofpeaks(j));
    Pp = polyfit(1:nofpeaks(j),dec_y{j}(1:(nofpeaks(j))),1) %1=linear
    reg_line{j}=polyval(Pp,pk{j});                           
    dec_slope(j)=Pp(1);                     %slope from linear regression
    si(j)=dec_slope(j)/(sqrt((dec_slope(j).^2+4*pi^2))); %damping ratio
end

%% PLOT regression on logaritmic decrement
for j=1:Nm
        subplot(Nm,1,j)
        plot(1:(length(upeak{j})-s(j)),dec_y{j}(1:end),'or','MarkerSize', 2)
        ylabel('k\delta_k')
        axis([0 length(upeak{j})-s(j)+4 0 max(dec_y{j}(1:end))+0.5])
        if j==Nm
        xlabel('k, number of peaks')
        end
        hold on
    plot(pk{j},reg_line{j},'Linewidth',1.2)
        [legh,objh,outh,outm] =legend(['R_s_' num2str(j) '_s_' num2str(j)])
        set(objh,'LineStyle','none');
        legend('boxoff')
   
end 
savefig(['ID' num2str(ID) 'logdecrement'])
%% PLOT Correlation functions with peaks
figure
suptitle('Correlation functions of sources')

for i=1:Nm
    subplot(Nm,1,i)
plot(1:K,Rrss(i,:))
legend('boxoff')
ylabel(['R_s_' num2str(i) '_s_' num2str(i)])
if i==Nm
xlabel('\tau')
end
hold on
plot(tpeak{i},upeak{i},'*r')
end

%% RESULTS
for i=1:length(f_Rrss)
fn_s(i)=f_Rrss(i)/sqrt(1-si((i))^2);
phi_unit(:,i)=Ar(:,i)/max(abs(Ar(:,i)));
end

modes.f=fn_s; %obtained frequencies
modes.m=phi_unit;%mode shapes, nomalized by unit
modes.xi=si;    %damping ratios
save(['ID' num2str(ID) '_SOBI'],'modes')
[fd num]=sort(f_Rrss);
